# -*- coding: utf-8 -*-

# Copyright (c) 2013-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

# pylint: disable=missing-docstring,attribute-defined-outside-init,protected-access


from unittest import TestCase

from pyramid.request import Request

from c2cgeoportal_geoportal import OgcproxyRoutePredicate


class TestMapserverproxyRoutePredicate(TestCase):

    predicate = OgcproxyRoutePredicate(None, None)

    def test_no_url(self):
        request = Request.blank("/test")
        self.assertFalse(self.predicate(None, request))

    def test_invalid(self):
        request = Request.blank("/test?url=http://askdljfhaskdjhfakldjs.com/")
        self.assertFalse(self.predicate(None, request))

    def test_google(self):
        request = Request.blank("/test?url=http://google.com/")
        self.assertTrue(self.predicate(None, request))

    def test_1234(self):
        request = Request.blank("/test?url=http://1.2.3.4/")
        self.assertTrue(self.predicate(None, request))

    def test_local(self):
        request = Request.blank("/test?url=http://127.2.3.4/")
        self.assertFalse(self.predicate(None, request))

    def test_10(self):
        request = Request.blank("/test?url=http://10.2.3.4/")
        self.assertFalse(self.predicate(None, request))

    def test_192_168(self):
        request = Request.blank("/test?url=http://192.168.3.4/")
        self.assertFalse(self.predicate(None, request))

    def test_172_16_0_0(self):
        request = Request.blank("/test?url=http://172.16.0.0/")
        self.assertFalse(self.predicate(None, request))

    def test_172_31_255_255(self):
        request = Request.blank("/test?url=http://172.31.255.255/")
        self.assertFalse(self.predicate(None, request))

    def test_172_15_255_255(self):
        request = Request.blank("/test?url=http://172.15.255.255/")
        self.assertTrue(self.predicate(None, request))

    def test_172_32_0_0(self):
        request = Request.blank("/test?url=http://172.32.0.0/")
        self.assertTrue(self.predicate(None, request))



# -*- coding: utf-8 -*-

# Copyright (c) 2013-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

# pylint: disable=missing-docstring,attribute-defined-outside-init,protected-access


from unittest import TestCase


class TestRasterViews(TestCase):

    def test_raster(self):
        from decimal import Decimal
        from pyramid.testing import DummyRequest
        from pyramid.httpexceptions import HTTPNotFound
        from c2cgeoportal_geoportal.views.raster import Raster

        request = DummyRequest()
        request.registry.settings = {
            "raster": {
                "dem1": {
                    "file": "/src/geoportal/tests/data/dem.shp",
                    "round": 0.1
                },
                "dem2": {
                    "file": "/src/geoportal/tests/data/dem.shp",
                    "round": 1
                },
                "dem3": {"file": "/src/geoportal/tests/data/dem.shp"}
            }
        }
        raster = Raster(request)

        request.params["lon"] = "565000"
        request.params["lat"] = "218000"
        result = raster.raster()
        self.assertEqual(result["dem1"], None)
        self.assertEqual(result["dem2"], None)
        self.assertEqual(result["dem3"], None)

        request.params["lon"] = "548000"
        request.params["lat"] = "216000"
        result = raster.raster()
        self.assertAlmostEqual(result["dem1"], Decimal("1171.6"))
        self.assertAlmostEqual(result["dem2"], Decimal("1172"))
        self.assertAlmostEqual(result["dem3"], Decimal("1171.62"))

        request.params["layers"] = "dem2"
        result = raster.raster()
        self.assertFalse("dem1" in result)
        self.assertFalse("dem3" in result)
        self.assertAlmostEqual(result["dem2"], Decimal("1172"))

        # test wrong layer name
        request.params["layers"] = "wrong"
        self.assertRaises(HTTPNotFound, raster.raster)

    def test_raster_angle(self):
        from decimal import Decimal
        from pyramid.testing import DummyRequest
        from c2cgeoportal_geoportal.views.raster import Raster

        request = DummyRequest()
        request.registry.settings = {
            "raster": {
                "dem5": {
                    "file": "/src/geoportal/tests/data/dem4.bt",
                    "type": "gdal",
                    "round": 0.01
                }
            }
        }
        raster = Raster(request)

        # Upper left
        request.params["lon"] = "547990.0"
        request.params["lat"] = "216009.1"
        result = raster.raster()
        self.assertEqual(result["dem5"], Decimal("1164.2"))
        request.params["lon"] = "547990.9"
        request.params["lat"] = "216010.0"
        result = raster.raster()
        self.assertEqual(result["dem5"], Decimal("1164.2"))

        # Lower right
        request.params["lon"] = "547996.0"
        request.params["lat"] = "216003.1"
        result = raster.raster()
        self.assertEqual(result["dem5"], Decimal("1180.77"))
        request.params["lon"] = "547996.9"
        request.params["lat"] = "216004.0"
        result = raster.raster()
        self.assertEqual(result["dem5"], Decimal("1180.77"))

        # Out
        request.params["lon"] = "547997.4"
        request.params["lat"] = "216003.5"
        result = raster.raster()
        self.assertEqual(result["dem5"], None)

    def test_raster_vrt(self):
        from decimal import Decimal
        from pyramid.testing import DummyRequest
        from c2cgeoportal_geoportal.views.raster import Raster

        request = DummyRequest()
        request.registry.settings = {
            "raster": {
                "dem6": {
                    "file": "/src/geoportal/tests/data/dem4.vrt",
                    "type": "gdal",
                    "round": 0.01
                }
            }
        }
        raster = Raster(request)

        # Upper left
        request.params["lon"] = "547990.4"
        request.params["lat"] = "216009.5"
        result = raster.raster()
        self.assertEqual(result["dem6"], Decimal("1164.2"))

    def test_absolute_path(self):
        import fiona
        with fiona.open("/src/geoportal/tests/data/dem_absolute.shp") as collection:
            tiles = [e for e in collection.filter(mask={
                "type": "Point",
                "coordinates": [548000, 216000],
            })]

        self.assertEqual(
            tiles[0]["properties"]["location"],
            "/home/sbrunner/regiogis/regiogis/c2cgeoportal/c2cgeoportal/tests/data/dem.bt")

    def test_profile_json(self):
        from decimal import Decimal
        from pyramid.testing import DummyRequest
        from pyramid.httpexceptions import HTTPNotFound
        from c2cgeoportal_geoportal.views.profile import Profile

        request = DummyRequest()
        request.registry.settings = {
            "raster": {
                "dem": {"file": "/src/geoportal/tests/data/dem.shp", "round": 4},
                "dem2": {"file": "/src/geoportal/tests/data/dem.shp", "round": 4}
            }
        }
        profile = Profile(request)

        request.params["nbPoints"] = "3"
        request.params["geom"] = '{"type":"LineString",' \
            '"coordinates":[[548009.5,215990],[547990,216009.5]]}'
        result = profile.json()
        self.assertEqual(len(result["profile"]), 4)
        self.assertAlmostEqual(result["profile"][0]["y"], 215990)
        self.assertAlmostEqual(result["profile"][0]["values"]["dem2"], None)
        self.assertAlmostEqual(result["profile"][0]["values"]["dem"], None)
        self.assertAlmostEqual(result["profile"][0]["dist"], Decimal("0.0"))
        self.assertAlmostEqual(result["profile"][0]["x"], 548009.5)
        self.assertAlmostEqual(result["profile"][1]["y"], 215996.5)
        self.assertAlmostEqual(result["profile"][1]["values"]["dem2"], 1181)
        self.assertAlmostEqual(result["profile"][1]["values"]["dem"], 1181)
        self.assertAlmostEqual(result["profile"][1]["dist"], Decimal("9.2"))
        self.assertAlmostEqual(result["profile"][1]["x"], 548003.0)
        self.assertAlmostEqual(result["profile"][2]["y"], 216003.0)
        self.assertAlmostEqual(result["profile"][2]["values"]["dem"], 1180)
        self.assertAlmostEqual(result["profile"][2]["values"]["dem2"], 1180)
        self.assertAlmostEqual(result["profile"][2]["dist"], Decimal("18.4"))
        self.assertAlmostEqual(result["profile"][2]["x"], 547996.5)
        self.assertAlmostEqual(result["profile"][3]["y"], 216009.5)
        self.assertAlmostEqual(result["profile"][3]["values"]["dem"], 1164)
        self.assertAlmostEqual(result["profile"][3]["values"]["dem2"], 1164)
        self.assertAlmostEqual(result["profile"][3]["dist"], Decimal("27.6"))
        self.assertAlmostEqual(result["profile"][3]["x"], 547990.0)

        request.params["layers"] = "dem"
        result = profile.json()
        self.assertEqual(len(result["profile"]), 4)
        self.assertAlmostEqual(result["profile"][0]["y"], 215990)
        self.assertAlmostEqual(result["profile"][0]["values"]["dem"], None)
        self.assertAlmostEqual(result["profile"][0]["dist"], Decimal("0.0"))
        self.assertAlmostEqual(result["profile"][0]["x"], 548009.5)
        self.assertAlmostEqual(result["profile"][1]["y"], 215996.5)
        self.assertAlmostEqual(result["profile"][1]["values"]["dem"], 1181)
        self.assertAlmostEqual(result["profile"][1]["dist"], Decimal("9.2"))
        self.assertAlmostEqual(result["profile"][1]["x"], 548003.0)
        self.assertAlmostEqual(result["profile"][2]["y"], 216003.0)
        self.assertAlmostEqual(result["profile"][2]["values"]["dem"], 1180)
        self.assertAlmostEqual(result["profile"][2]["dist"], Decimal("18.4"))
        self.assertAlmostEqual(result["profile"][2]["x"], 547996.5)
        self.assertAlmostEqual(result["profile"][3]["y"], 216009.5)
        self.assertAlmostEqual(result["profile"][3]["values"]["dem"], 1164)
        self.assertAlmostEqual(result["profile"][3]["dist"], Decimal("27.6"))
        self.assertAlmostEqual(result["profile"][3]["x"], 547990.0)

        # test length = 0
        request.params["geom"] = '{"type":"LineString",' \
            '"coordinates":[[548000,216000]]}'
        result = profile.json()
        self.assertEqual(len(result["profile"]), 1)
        self.assertAlmostEqual(result["profile"][0]["y"], 216000)
        self.assertAlmostEqual(result["profile"][0]["values"]["dem"], 1172)
        self.assertAlmostEqual(result["profile"][0]["dist"], Decimal("0.0"))
        self.assertAlmostEqual(result["profile"][0]["x"], 548000)

        # test cur_nb_points < 1
        request.params["geom"] = '{"type":"LineString",' \
            '"coordinates":[[548000,216000],[548001,216001],[548009,216009]]}'
        result = profile.json()
        self.assertEqual(len(result["profile"]), 5)
        self.assertAlmostEqual(result["profile"][0]["y"], 216000)
        self.assertAlmostEqual(result["profile"][0]["values"]["dem"], 1172)
        self.assertAlmostEqual(result["profile"][0]["dist"], Decimal("0.0"))
        self.assertAlmostEqual(result["profile"][0]["x"], 548000)
        self.assertAlmostEqual(result["profile"][1]["y"], 216001.0)
        self.assertAlmostEqual(result["profile"][1]["values"]["dem"], 1167)
        self.assertAlmostEqual(result["profile"][1]["dist"], Decimal("1.4"))
        self.assertEqual(result["profile"][1]["x"], 548001.0)
        self.assertAlmostEqual(result["profile"][2]["y"], 216003.66666666666)
        self.assertAlmostEqual(result["profile"][2]["values"]["dem"], 1155)
        self.assertAlmostEqual(result["profile"][2]["dist"], Decimal("5.2"))
        self.assertAlmostEqual(result["profile"][2]["x"], 548003.6666666666)
        self.assertAlmostEqual(result["profile"][3]["y"], 216006.33333333334)
        self.assertAlmostEqual(result["profile"][3]["values"]["dem"], 1154)
        self.assertAlmostEqual(result["profile"][3]["dist"], Decimal("9"))
        self.assertAlmostEqual(result["profile"][3]["x"], 548006.3333333334)
        self.assertAlmostEqual(result["profile"][4]["y"], 216009.0)
        self.assertAlmostEqual(result["profile"][4]["values"]["dem"], 1158)
        self.assertAlmostEqual(result["profile"][4]["dist"], Decimal("12.7"))
        self.assertAlmostEqual(result["profile"][4]["x"], 548009.0)

        # test wrong layer name
        request.params["layers"] = "wrong"
        self.assertRaises(HTTPNotFound, profile.json)

    def test_profile_csv(self):
        from pyramid.testing import DummyRequest
        from c2cgeoportal_geoportal.views.profile import Profile

        request = DummyRequest()
        request.registry.settings = {
            "raster": {
                "dem": {"file": "/src/geoportal/tests/data/dem.shp", "round": 1},
                "dem2": {"file": "/src/geoportal/tests/data/dem.shp", "round": 1},
                "dem4": {"file": "/src/geoportal/tests/data/dem4.shp", "round": 1}
            }
        }
        profile = Profile(request)

        request.params["nbPoints"] = "3"
        request.params["geom"] = '{"type":"LineString",' \
            '"coordinates":[[548009.5,215990],[547990,216009.5]]}'
        response = profile.csv()
        self.assertEqual(response.body.decode("utf-8"), """distance,dem,dem2,dem4,x,y
0.0,-9999,-9999,-9999,548009.5,215990.0
9.2,1181,1181,-9999,548003.0,215996.5
18.4,1180,1180,-9999,547996.5,216003.0
27.6,1164,1164,1164,547990.0,216009.5""")

        request.params["layers"] = "dem"
        response = profile.csv()
        self.assertEqual(response.body.decode("utf-8"), """distance,dem,x,y
0.0,-9999,548009.5,215990.0
9.2,1181,548003.0,215996.5
18.4,1180,547996.5,216003.0
27.6,1164,547990.0,216009.5""")



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

# pylint: disable=missing-docstring,attribute-defined-outside-init,protected-access


import json
import decimal
from unittest import TestCase
from pyramid import testing


class TestDecimalJSON(TestCase):
    @staticmethod
    def _call_fut(**kwargs):
        from c2cgeoportal_geoportal import DecimalJSON
        fake_info = {}
        return DecimalJSON(**kwargs)(fake_info)

    def test_decimal(self):
        renderer = self._call_fut()
        value = {
            "str": "an str",
            "int": 1,
            "dec": decimal.Decimal("1.2")
        }
        request = testing.DummyRequest()
        result = renderer(value, {"request": request})
        self.assertEqual(
            json.loads(result),
            {
                "int": 1,
                "dec": 1.2,
                "str": "an str"
            }
        )
        self.assertEqual(request.response.content_type, "application/json")



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


"""Pyramid application test package
"""

import os
from pyramid.testing import DummyRequest


def create_dummy_request(additional_settings=None, *args, **kargs):
    if additional_settings is None:
        additional_settings = {}
    request = DummyRequest(*args, **kargs)
    request.registry.settings = {
        "available_locale_names": ["en", "fr", "de"],
        "default_locale_name": "fr",
        "default_max_age": 1000,
        "package": "package_for_test"
    }
    request.registry.settings.update(additional_settings)
    request.is_valid_referer = True
    return request


def load_binfile(file_name):
    with open(os.path.join("/src/geoportal", file_name), "rb") as file_:
        return file_.read()


def load_file(file_name):
    with open(os.path.join("/src/geoportal", file_name)) as file_:
        return file_.read()



# -*- coding: utf-8 -*-

# Copyright (c) 2013-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

# pylint: disable=missing-docstring,attribute-defined-outside-init,protected-access


from unittest import TestCase


class TestFunctionalities(TestCase):
    def test_get_config_functionality(self):
        from c2cgeoportal_geoportal.lib import get_types_map
        from c2cgeoportal_geoportal.lib import functionality

        types = get_types_map([{"name": "func"}])

        class Registry:
            settings = {}

        class Request:
            registry = Registry()

        request = Request()
        request.registry.settings = {
            "functionalities": {
                "registered": {
                    "func": 10
                },
                "anonymous": {
                    "func": 20
                }
            }
        }
        functionality.FUNCTIONALITIES_TYPES = None
        errors = set()

        f = functionality._get_config_functionality("func", True, types, request, errors)
        self.assertEqual(errors, set())
        self.assertEqual(f, [10])

        f = functionality._get_config_functionality("func", False, types, request, errors)
        self.assertEqual(errors, set())
        self.assertEqual(f, [20])

        request.registry.settings = {
            "functionalities": {
                "registered": {
                    "not_func": 10
                },
                "anonymous": {
                    "func": 20
                }
            }
        }
        functionality.FUNCTIONALITIES_TYPES = None
        f = functionality._get_config_functionality("func", True, types, request, errors)
        self.assertEqual(errors, set())
        self.assertEqual(f, [20])

        request.registry.settings = {
            "functionalities": {
                "registered": {
                    "func": 10
                },
                "anonymous": {
                    "not_func": 20
                }
            }
        }
        functionality.FUNCTIONALITIES_TYPES = None
        f = functionality._get_config_functionality("func", False, types, request, errors)
        self.assertEqual(errors, set())
        self.assertEqual(f, [])



# -*- coding: utf-8 -*-

# Copyright (c) 2013-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


getfeature = """
<wfs:GetFeature xmlns:wfs="http://www.opengis.net/wfs" service="WFS" version="1.1.0" xsi:schemaLocation="http://www.opengis.net/wfs http://schemas.opengis.net/wfs/1.1.0/wfs.xsd" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
<wfs:Query typeName="feature:grundstueck" srsName="EPSG:2056" xmlns:feature="http://mapserver.gis.umn.edu/mapserver">
<ogc:Filter xmlns:ogc="http://www.opengis.net/ogc">
<ogc:PropertyIsLike matchCase="false" wildCard="*" singleChar="." escapeChar="!">
<ogc:PropertyName>nummer</ogc:PropertyName>
<ogc:Literal>10*</ogc:Literal>
</ogc:PropertyIsLike>
</ogc:Filter>
</wfs:Query>
</wfs:GetFeature>

"""
feature = """
    <gml:featureMember>
      <ms:grundstueck>
        <gml:boundedBy>
            <gml:Envelope srsName="EPSG:2056">
                <gml:lowerCorner>2626901.051818 1258035.790009</gml:lowerCorner>
                <gml:upperCorner>2627050.862856 1258132.841364</gml:upperCorner>
            </gml:Envelope>
        </gml:boundedBy>
        <ms:msGeometry>
          <gml:LineString srsName="EPSG:2056">
            <gml:posList srsDimension="2">2627033.201116 1258103.390372 2627034.048142 1258105.737388 2627010.821109 1258118.506850 2626985.111074 1258132.841364 2626980.135958 1258123.622322 2626978.010913 1258120.089309 2626966.170890 1258126.005538 2626949.985629 1258108.760552 2626924.919220 1258081.422566 2626910.187979 1258065.386575 2626901.051818 1258054.063564 2626935.224905 1258039.509934 2626956.098017 1258037.068626 2626971.167108 1258036.400415 2627000.949294 1258035.790009 2627018.708458 1258041.255835 2627029.967583 1258047.114753 2627048.056822 1258060.580669 2627050.862856 1258062.337652 2627048.942861 1258064.236700 2627036.107888 1258076.303014 2627023.360917 1258088.497329 2627028.596025 1258096.640354 2627033.201116 1258103.390372 </gml:posList>
          </gml:LineString>
        </ms:msGeometry>
        <ms:gs_id>1676545</ms:gs_id>
        <ms:lsn_oid>1510175178</ms:lsn_oid>
        <ms:nummer>1071</ms:nummer>
        <ms:gueltigkeit>rechtskräftig</ms:gueltigkeit>
        <ms:art>Liegenschaft</ms:art>
        <ms:gemeinde_id_bfs>2861</ms:gemeinde_id_bfs>
        <ms:meta_id>1510</ms:meta_id>
        <ms:flaechenmass>8774</ms:flaechenmass>
        <ms:nummer_m_deko>1071</ms:nummer_m_deko>
        <ms:nbident>BL0200002861</ms:nbident>
        <ms:vollstaendigkeit>vollständig</ms:vollstaendigkeit>
        <ms:datenherr>Jermann</ms:datenherr>
        <ms:mut_nummer>pn18</ms:mut_nummer>
      </ms:grundstueck>
    </gml:featureMember>
    """

featurecollection_outlimit = """
<wfs:FeatureCollection xmlns:ms="http://mapserver.gis.umn.edu/mapserver" xmlns:ogc="http://www.opengis.net/ogc" xmlns:gml="http://www.opengis.net/gml" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:wfs="http://www.opengis.net/wfs" xsi:schemaLocation="http://mapserver.gis.umn.edu/mapserver http://c2cpc29.camptocamp.com/sbrunner/mapserv?SERVICE=WFS&amp;VERSION=1.1.0&amp;REQUEST=DescribeFeatureType&amp;TYPENAME=feature:grundstueck&amp;OUTPUTFORMAT=text/xml;%20subtype=gml/3.1.1  http://www.opengis.net/wfs http://schemas.opengis.net/wfs/1.1.0/wfs.xsd">
    <gml:boundedBy>
      <gml:Envelope srsName="EPSG:2056">
          <gml:lowerCorner>2595270.118588 1244096.257242</gml:lowerCorner>
          <gml:upperCorner>2638409.063753 1267658.751429</gml:upperCorner>
      </gml:Envelope>
    </gml:boundedBy>
    """ + feature * 205 + """
</wfs:FeatureCollection>
    """

featurecollection_inlimit = """
<wfs:FeatureCollection xmlns:ms="http://mapserver.gis.umn.edu/mapserver" xmlns:ogc="http://www.opengis.net/ogc" xmlns:gml="http://www.opengis.net/gml" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:wfs="http://www.opengis.net/wfs" xsi:schemaLocation="http://mapserver.gis.umn.edu/mapserver http://c2cpc29.camptocamp.com/sbrunner/mapserv?SERVICE=WFS&amp;VERSION=1.1.0&amp;REQUEST=DescribeFeatureType&amp;TYPENAME=feature:grundstueck&amp;OUTPUTFORMAT=text/xml;%20subtype=gml/3.1.1  http://www.opengis.net/wfs http://schemas.opengis.net/wfs/1.1.0/wfs.xsd">
    <gml:boundedBy>
      <gml:Envelope srsName="EPSG:2056">
          <gml:lowerCorner>2595270.118588 1244096.257242</gml:lowerCorner>
          <gml:upperCorner>2638409.063753 1267658.751429</gml:upperCorner>
      </gml:Envelope>
    </gml:boundedBy>
    """ + feature * 199 + """
</wfs:FeatureCollection>
    """



# -*- coding: utf-8 -*-

# Copyright (c) 2012-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

# pylint: disable=missing-docstring,attribute-defined-outside-init,protected-access


from unittest import TestCase
from pyramid import testing
from pyramid.testing import DummyRequest

from c2c.template.config import config
import c2cgeoportal_geoportal
from c2cgeoportal_geoportal import call_hook, set_user_validator, \
    default_user_validator, create_get_user_from_request, is_valid_referer


class TestIncludeme(TestCase):

    config = None

    def setup_method(self, _):
        # the c2cgeoportal includeme function requires a number
        # of settings
        self.config = testing.setUp(settings={
            "sqlalchemy.url": "postgresql://www-data:www-data@db:5432/geomapfish_tests",
            "srid": 3857,
            "schema": "main",
            "schema_static": "main_static",
            "default_max_age": 86400,
            "app.cfg": "/src/geoportal/tests/config.yaml",
            "package": "c2cgeoportal",
            "enable_admin_interface": False,
        })

    def test_set_user_validator_directive(self):
        self.config.include(c2cgeoportal_geoportal.includeme)
        self.assertTrue(
            self.config.set_user_validator.__func__.__docobj__ is
            set_user_validator
        )

    def test_default_user_validator(self):
        self.config.include(c2cgeoportal_geoportal.includeme)
        self.assertEqual(
            self.config.registry.validate_user,
            default_user_validator
        )

    def test_user_validator_overwrite(self):
        self.config.include(c2cgeoportal_geoportal.includeme)

        def custom_validator(username, password):
            del username  # Unused
            del password  # Unused
            return False  # pragma: no cover
        self.config.set_user_validator(custom_validator)
        self.assertEqual(self.config.registry.validate_user, custom_validator)


class TestReferer(TestCase):
    """
    Check that accessing something with a bad HTTP referer is equivalent to a
    not authenticated query.
    """
    BASE1 = "http://example.com/app"
    BASE2 = "http://friend.com/app2"
    SETTINGS = {"authorized_referers": [
        BASE1,
        BASE2
    ]}
    USER = "toto"

    def _get_user(self, to, ref, method="GET"):
        class MockRequest:
            def __init__(self, to, ref, method):
                self.path_qs = to
                self.referer = ref
                self.user_ = TestReferer.USER
                self.method = method

        config._config = {
            'schema': 'main',
            'schema_static': 'main_static',
            'srid': 21781
        }
        get_user = create_get_user_from_request(self.SETTINGS)
        return get_user(MockRequest(to=to, ref=ref, method=method))

    def test_match_url(self):
        def match(reference, value, expected):
            r = DummyRequest()
            r.referer = value
            self.assertEqual(is_valid_referer(r, {
                "authorized_referers": [reference]
            }), expected)

        match("http://example.com/app/", "http://example.com/app?k=v", True)
        match("http://example.com/app/", "http://example.com/app?k=v#link", True)
        match("http://example.com/app/", "http://example.com/app#link", True)
        match("http://example.com/app/", "http://example.com/app", True)
        match("http://example.com/app", "http://example.com/app/", True)
        match("http://example.com/app", "http://example.com/app/x/y", True)
        match("http://example.com", "http://example.com/app/x/y", True)
        match("http://example.com", "http://other.com", False)
        match("http://example.com", "https://example.com", False)
        match("http://example.com/app", "http://example.com/", False)
        match("http://example.com", "http://example.com.bad.org/app/x/y", False)

    def test_positive(self):
        self.assertEqual(
            self._get_user(to=self.BASE1 + "/1", ref=self.BASE1), self.USER)
        self.assertEqual(
            self._get_user(to=self.BASE1 + "/2", ref=self.BASE1 + "/3"), self.USER)
        self.assertEqual(
            self._get_user(to=self.BASE1 + "/4", ref=self.BASE2 + "/5"), self.USER)

    def test_no_ref(self):
        self.assertEqual(self._get_user(to=self.BASE1, ref=None), self.USER)
        self.assertIsNone(self._get_user(to=self.BASE1, ref=""))

        self.assertEqual(self._get_user(to=self.BASE1, ref=None, method="POST"), self.USER)
        self.assertIsNone(self._get_user(to=self.BASE1, ref="", method="POST"))

    def test_bad_ref(self):
        self.assertIsNone(self._get_user(
            to=self.BASE1,
            ref="http://bad.com/hacker"))


def hook(tracer):
    tracer["called"] = True


class TestHooks(TestCase):
    settings = {
        "hooks": {
            "test": "tests.test_init.hook",
            "bad": "c2cgeoportal_geoportal.not_here"
        }
    }

    def test_existing(self):
        tracer = {"called": False}
        call_hook(self.settings, "test", tracer)
        self.assertTrue(tracer["called"])

    def test_no_hook(self):
        call_hook(self.settings, "test2")

    @staticmethod
    def test_no_hooks():
        call_hook({}, "test")

    def test_bad_hook(self):
        self.assertRaises(AttributeError, call_hook, self.settings, "bad")


class TestInit(TestCase):
    def test_add_url_params(self):
        from c2cgeoportal_geoportal.lib import add_url_params
        from urllib.parse import urlparse, parse_qs
        params = {"Name": "Bob", "Age": 18, "Nationality": "Việt Nam"}
        result = add_url_params("http://test/", params)
        presult = urlparse(result)
        self.assertEqual(presult.scheme, "http")
        self.assertEqual(presult.netloc, "test")
        self.assertEqual(presult.path, "/")
        self.assertEqual(presult.params, "")
        self.assertEqual(presult.fragment, "")
        self.assertEqual(parse_qs(presult.query), {
            "Name": ["Bob"],
            "Age": ["18"],
            "Nationality": ["Việt Nam"],
        })



# -*- coding: utf-8 -*-

# Copyright (c) 2018-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import logging
from typing import Any, Dict, List

from c2cwsgiutils import broadcast
from c2cwsgiutils.auth import auth_view
from pympler import asizeof, muppy
from pyramid.view import view_config

from c2cgeoportal_geoportal.views import entry, raster

LOG = logging.getLogger(__name__)


@view_config(route_name="memory", renderer="fast_json")
def memory(request):
    auth_view(request)
    return _memory(
        with_others=request.params.get("with_others", "false").lower() in ["1", "on", "true"],
        with_all=request.params.get("with_all", "false").lower() in ["1", "on", "true"],
        with_repr=request.params.get("with_repr", "false").lower() in ["1", "on", "true"],
    )


@broadcast.decorator(expect_answers=True, timeout=110)
def _memory(with_others: bool, with_all: bool, with_repr: bool) -> Dict[str, Any]:
    wms_capabilities_cache = entry.Entry.server_wms_capabilities
    wfs_feature_type_cache = entry.Entry.server_wfs_feature_type
    raster_data = raster.Raster.data
    result = {
        "parsed_wms_capabilities_cache_by_ogcserver_id": {
            id: _get_size(wms_capabilities_cache[id]) / 1024 / 1024 for id in wms_capabilities_cache
        },
        "parsed_wfs_feature_type_cache_by_ogcserver_id": {
            id: _get_size(wfs_feature_type_cache[id]) / 1024 / 1024 for id in wfs_feature_type_cache
        },
        "raster_data": {
            id: _get_size(raster_data[id]) / 1024 / 1024 for id in raster_data
        },
    }  # type: Dict[str, Any]

    if with_others:
        result['others'] = _get_used_memory(with_all, with_repr)

    return result


def _filter(module_: str, with_all: bool) -> bool:
    return with_all or (
        module_ not in ("builtins", "<unknown>")
        and not module_.startswith("_")
        and not module_.startswith("pkg_resources")
    )


def _get_size(obj: Any) -> int:
    try:
        return asizeof.asizeof(obj)
    except Exception:
        LOG.exception("Unable to get the size")
        return 0


def _order_by_size(elements: Dict[str, Any]) -> List[Dict[str, Any]]:
    for name, elem in elements.items():
        if 'children' in elem:
            elem['children'] = _order_by_size(elem['children'])

    return sorted(elements.values(), key=lambda elem: elem['size'])


def _get_used_memory(with_all: bool, with_repr: bool) -> List[Dict[str, Any]]:
    all_objects = [(
        ".".join((type(o).__module__, type(o).__name__)),
        o,
        _get_size(o)
    ) for o in muppy.get_objects() if _filter(type(o).__module__, with_all)]
    result = {}  # type: Dict[str, Any]
    for name, obj, size in all_objects:
        path = name.split(".")
        _fill_path(result, path, obj, size, with_repr)

    return _order_by_size(result)


def _fill_path(base: Dict[str, Any], path: List[str], object_: Any, size: int, with_repr: bool):
    if len(path) == 1:
        base.setdefault(path[0], {
            "name": path[0],
            "size": 0,
            "numbers": 0,
        })
        obj = base[path[0]]
        obj["size"] += size
        obj["numbers"] += 1
        if with_repr:
            obj.setdefault("objects", [])
            obj["objects"].append((repr(object_), size))
    else:
        base.setdefault(path[0], {
            "name": path[0],
            "size": 0,
            "numbers": 0,
            "children": {},
        })
        obj = base[path[0]]
        obj["size"] += size
        obj["numbers"] += 1
        _fill_path(obj["children"], path[1:], object_, size, with_repr)



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import logging
from typing import Any, Dict  # noqa, pylint: disable=unused-import

from pyramid.response import Response
from pyramid.view import view_config
from pyramid.request import Request

from c2cgeoportal_geoportal.lib.caching import get_region, NO_CACHE, PUBLIC_CACHE, PRIVATE_CACHE
from c2cgeoportal_geoportal.lib.functionality import get_mapserver_substitution_params
from c2cgeoportal_geoportal.lib.filter_capabilities import filter_capabilities
from c2cgeoportal_geoportal.views.ogcproxy import OGCProxy
from c2cgeoportal_commons.models import main, static

cache_region = get_region()
log = logging.getLogger(__name__)


class MapservProxy(OGCProxy):

    params = {}  # type: Dict[str, str]

    def __init__(self, request: Request) -> None:
        OGCProxy.__init__(self, request)
        self.user = self.request.user

    @view_config(route_name="mapserverproxy")
    def proxy(self) -> Response:
        if self.user is not None:
            # We have a user logged in. We need to set group_id and
            # possible layer_name in the params. We set layer_name
            # when either QUERY_PARAMS or LAYERS is set in the
            # WMS params, i.e. for GetMap and GetFeatureInfo
            # requests. For GetLegendGraphic requests we do not
            # send layer_name, but MapServer should not use the DATA
            # string for GetLegendGraphic.

            roles = self.user.roles
            if len(roles):
                if self.ogc_server.auth == main.OGCSERVER_AUTH_STANDARD:
                    self.params["role_ids"] = \
                        "-1" if len(roles) == 0 else ",".join([str(r.id) for r in roles])

                    # In some application we want to display the features owned by a user
                    # than we need his id.
                    self.params["user_id"] = self.user.id  # pragma: no cover
            else:  # pragma nocover
                log.warning("The user '%s' has no role", self.user.username)

        # do not allows direct variable substitution
        for k in list(self.params.keys()):
            if k[:2].capitalize() == "S_":
                log.warning("Direct substitution not allowed ({0!s}={1!s}).".format(k, self.params[k]))
                del self.params[k]

        # add functionalities params
        self.params.update(get_mapserver_substitution_params(self.request))

        # get method
        method = self.request.method

        # we want the browser to cache GetLegendGraphic and
        # DescribeFeatureType requests
        use_cache = False

        if method == "GET":
            # For GET requests, params are added only if the self.request
            # parameter is actually provided.
            if "request" not in self.lower_params:
                self.params = {}  # pragma: no cover
            else:
                if self.ogc_server.type != main.OGCSERVER_TYPE_QGISSERVER or \
                        "user_id" not in self.params:

                    use_cache = self.lower_params["request"] in (
                        "getlegendgraphic",
                    )

                    # no user_id and role_id or cached queries
                    if use_cache and "user_id" in self.params:
                        del self.params["user_id"]
                    if use_cache and "role_ids" in self.params:
                        del self.params["role_ids"]

            if "service" in self.lower_params and self.lower_params["service"] == "wfs":
                _url = self._get_wfs_url()
            else:
                _url = self._get_wms_url()
        else:
            # POST means WFS
            _url = self._get_wfs_url()

        cache_control = PRIVATE_CACHE
        if method == "GET" and \
                "service" in self.lower_params and \
                self.lower_params["service"] == "wms":
            if self.lower_params["request"] in ("getmap", "getfeatureinfo"):
                cache_control = NO_CACHE
            elif self.lower_params["request"] == "getlegendgraphic":
                cache_control = PUBLIC_CACHE
        elif method == "GET" and \
                "service" in self.lower_params and \
                self.lower_params["service"] == "wfs":
            if self.lower_params["request"] == "getfeature":
                cache_control = NO_CACHE
        elif method != "GET":
            cache_control = NO_CACHE

        headers = self._get_headers()
        # Add headers for Geoserver
        if self.ogc_server.auth == main.OGCSERVER_AUTH_GEOSERVER and \
                self.user is not None:
            headers["sec-username"] = self.user.username
            headers["sec-roles"] = ";".join([r.name for r in roles])

        response = self._proxy_callback(
            self.user, cache_control,
            url=_url, params=self.params, cache=use_cache, headers=headers, body=self.request.body
        )

        if self.lower_params.get("request") == "getmap" and \
                not response.content_type.startswith('image/') and \
                response.status_code < 400:
            response.status_code = 400

        return response

    def _proxy_callback(
        self, user: static.User, cache_control: int, url: str, params: dict, **kwargs: Any
    ) -> Response:
        callback = params.get("callback")
        if callback is not None:
            del params["callback"]
        response = self._proxy(url=url, params=params, **kwargs)

        content = response.content
        if self.lower_params.get("request") == "getcapabilities":
            content = filter_capabilities(
                response.text, user, self.lower_params.get("service") == "wms",
                url,
                self.request.headers,
                self.request
            ).encode("utf-8")

        if callback is not None:
            content_type = "application/javascript"
            # escape single quotes in the JavaScript string
            content = "{}('{}');".format(callback, " ".join(
                content.decode("utf-8").replace("'", "\\'").splitlines()
            )).encode("utf-8")
        else:
            content_type = response.headers["Content-Type"]

        return self._build_response(
            response, content, cache_control, "mapserver",
            content_type=content_type
        )



# -*- coding: utf-8 -*-

# Copyright (c) 2018-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


from c2cgeoportal_commons.models import static
from c2cgeoportal_geoportal.lib import caching


cache_region = caching.get_region()


def _get_layers_query(user: static.User, what):
    from c2cgeoportal_commons.models import DBSession, main

    assert user is not None

    q = DBSession.query(what)
    q = q.join(main.Layer.restrictionareas)
    q = q.join(main.RestrictionArea.roles)
    q = q.filter(main.Role.id.in_([r.id for r in user.roles]))

    return q


def get_protected_layers_query(user: static.User, ogc_server_ids, what=None):
    from c2cgeoportal_commons.models import main

    assert user is not None

    q = _get_layers_query(user, what)
    q = q.filter(main.Layer.public.is_(False))
    if ogc_server_ids is not None:
        q = q.join(main.LayerWMS.ogc_server)
        q = q.filter(main.OGCServer.id.in_(ogc_server_ids))
    return q


def get_writable_layers_query(user: static.User, ogc_server_ids):
    from c2cgeoportal_commons.models import main

    assert user is not None

    q = _get_layers_query(user, main.LayerWMS)
    return q \
        .filter(main.RestrictionArea.readwrite.is_(True)) \
        .join(main.LayerWMS.ogc_server) \
        .filter(main.OGCServer.id.in_(ogc_server_ids))


def get_protected_layers(user: static.User, ogc_server_ids):
    from c2cgeoportal_commons.models import DBSession, main

    if user is None:
        return {}

    q = get_protected_layers_query(user, ogc_server_ids, what=main.LayerWMS)
    results = q.all()
    DBSession.expunge_all()
    return {r.id: r for r in results}


def get_writable_layers(user: static.User, ogc_server_ids):
    from c2cgeoportal_commons.models import DBSession

    if user is None:
        return {}

    q = get_writable_layers_query(user, ogc_server_ids)
    results = q.all()
    DBSession.expunge_all()
    return {r.id: r for r in results}


@cache_region.cache_on_arguments()
def get_private_layers(ogc_server_ids):
    from c2cgeoportal_commons.models import DBSession, main

    q = DBSession.query(main.LayerWMS) \
        .filter(main.LayerWMS.public.is_(False)) \
        .join(main.LayerWMS.ogc_server) \
        .filter(main.OGCServer.id.in_(ogc_server_ids))
    results = q.all()
    DBSession.expunge_all()
    return {r.id: r for r in results}



# -*- coding: utf-8 -*-

# Copyright (c) 2015-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

# pylint: disable=missing-docstring,attribute-defined-outside-init,protected-access


from unittest import TestCase

from pyramid.request import Request, Response

from c2cgeoportal_geoportal.lib.caching import set_cors_headers, CORS_METHODS


class TestSetCorsHeaders(TestCase):
    ORIGIN1 = "http://www.example.com"
    ORIGIN2 = "http://www.friend.com"
    MAX_AGE = "1234"
    SETTINGS = {
        "access_control_allow_origin": [ORIGIN1, ORIGIN2],
        "access_control_max_age": MAX_AGE
    }

    @staticmethod
    def _do(method, headers, credentials=False, settings=SETTINGS):
        req = Request({}, method=method, headers=headers)
        req.response = Response()
        set_cors_headers(settings, req, "foo", credentials, req.response)
        return req.response.headers

    def _assert_headers(self, actual, expected):
        # need to add the headers automatically added by Request
        base_headers = {
            "Content-Type": "text/html; charset=UTF-8",
            "Content-Length": "0"
        }
        expected = expected.copy()
        expected.update(base_headers)
        self.assertEqual(actual, expected)

    def test_simple(self):
        """
        Tests specified in http://www.w3.org/TR/cors/#resource-requests
        """
        # 1. If the Origin header is not present terminate this set of steps.
        #    The request is outside the scope of this specification.
        self._assert_headers(self._do("POST", {}), {})

        # 2. If the value of the Origin header is not a case-sensitive match for
        #    any of the values in list of origins, do not set any additional
        #    headers and terminate this set of steps.
        self._assert_headers(self._do("POST", {"Origin": "http://foe.com"}), {})

        # 3. If the resource supports credentials add a single
        #    Access-Control-Allow-Origin header, with the value of the Origin
        #    header as value, and add a single Access-Control-Allow-Credentials
        #    header with the case-sensitive string "true" as value.
        self._assert_headers(self._do("POST", {
            "Origin": self. ORIGIN2
        }, credentials=True), {
            "Access-Control-Allow-Origin": self.ORIGIN2,
            "Access-Control-Allow-Credentials": "true",
            "Access-Control-Allow-Methods": CORS_METHODS,
            "Vary": set(["Origin"]),
        })

        # 3. Otherwise, add a single Access-Control-Allow-Origin header, with
        #    either the value of the Origin header or the string "*" as value.
        self._assert_headers(self._do("POST", {"Origin": self. ORIGIN2}), {
            "Access-Control-Allow-Origin": self.ORIGIN2,
            "Access-Control-Allow-Methods": CORS_METHODS,
            "Vary": set(["Origin"]),
        })

        # 4. If the list of exposed headers is not empty add one or more
        #    Access-Control-Expose-Headers headers, with as values the header
        #    field names given in the list of exposed headers.
        # Not implemented

    def test_preflight(self):
        """
        Tests specified in
        http://www.w3.org/TR/cors/#resource-preflight-requests
        """
        # 1. If the Origin header is not present terminate this set of steps.
        #    The request is outside the scope of this specification.
        self._assert_headers(self._do("OPTIONS", {}), {})

        # 2. If the value of the Origin header is not a case-sensitive match for
        #    any of the values in list of origins do not set any additional
        #    headers and terminate this set of steps.
        self._assert_headers(self._do("OPTIONS", {"Origin": "http://foe.com"}),
                             {})

        # 3. If there is no Access-Control-Request-Method header or if parsing
        #    failed, do not set any additional headers and terminate this set
        #    of steps. The request is outside the scope of this specification.
        self._assert_headers(self._do("OPTIONS", {
            "Origin": self.ORIGIN1}), {})

        # 4. If there are no Access-Control-Request-Headers headers let header
        #    field-names be the empty list.
        self._assert_headers(self._do("OPTIONS", {
            "Origin": self.ORIGIN1,
            "Access-Control-Request-Method": "GET"
        }), {
            "Access-Control-Allow-Origin": self.ORIGIN1,
            "Access-Control-Max-Age": self.MAX_AGE,
            "Access-Control-Allow-Methods": CORS_METHODS,
            "Vary": set(["Origin"]),
        })

        # 5. If method is not a case-sensitive match for any of the values in
        #    list of methods do not set any additional headers and terminate
        #    this set of steps.
        # Not implemented

        # 6. If any of the header field-names is not a ASCII case-insensitive
        #    match for any of the values in list of headers do not set any
        #    additional headers and terminate this set of steps.
        # Not implemented

        # 7. If the resource supports credentials add a single
        #    Access-Control-Allow-Origin header, with the value of the Origin
        #    header as value, and add a single Access-Control-Allow-Credentials
        #    header with the case-sensitive string "true" as value.
        self._assert_headers(self._do("OPTIONS", {
            "Origin": self.ORIGIN1,
            "Access-Control-Request-Method": "GET"
        }, credentials=True), {
            "Access-Control-Allow-Origin": self.ORIGIN1,
            "Access-Control-Allow-Credentials": "true",
            "Access-Control-Max-Age": self.MAX_AGE,
            "Access-Control-Allow-Methods": CORS_METHODS,
            "Vary": set(["Origin"]),
        })

        # 8. Optionally add a single Access-Control-Max-Age header with as value
        #    the amount of seconds the user agent is allowed to cache the result
        #    of the request.
        # Already tested

        # 9. Add one or more Access-Control-Allow-Methods headers consisting of
        #    (a subset of) the list of methods.
        # Already tested

        # 10. Add one or more Access-Control-Allow-Headers headers consisting of
        #     (a subset of) the list of headers.
        self._assert_headers(self._do("OPTIONS", {
            "Origin": self.ORIGIN1,
            "Access-Control-Request-Method": "GET",
            "Access-Control-Request-Headers": "X-Foo, X-Bar"
        }), {
            "Access-Control-Allow-Origin": self.ORIGIN1,
            "Access-Control-Max-Age": self.MAX_AGE,
            "Access-Control-Allow-Methods": CORS_METHODS,
            "Access-Control-Allow-Headers": "X-Foo, X-Bar",
            "Vary": set(["Origin"]),
        })

    def test_not_configured(self):
        # If the service is not configured, then no CORS head.
        self._assert_headers(
            self._do("GET", {"Origin": self. ORIGIN1}, settings=None), {})

    def test_match_all(self):
        settings = {
            "access_control_allow_origin": [self.ORIGIN1, "*"],
            "access_control_max_age": self.MAX_AGE
        }

        # An origin included in the access_control_allow_origin list is OK with
        # credentials
        self._assert_headers(self._do("POST", {
            "Origin": self.ORIGIN1
        }, credentials=True, settings=settings), {
            "Access-Control-Allow-Origin": self.ORIGIN1,
            "Access-Control-Allow-Credentials": "true",
            "Access-Control-Allow-Methods": CORS_METHODS,
            "Vary": set(["Origin"]),
        })

        # An origin not included in the access_control_allow_origin is allowed,
        # but without credentials
        self._assert_headers(self._do("POST", {
            "Origin": "http://guest.com"
        }, credentials=True, settings=settings), {
            "Access-Control-Allow-Origin": "*",
            "Access-Control-Allow-Methods": CORS_METHODS,
            "Vary": set(["Origin"]),
        })



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

import logging
from typing import Dict  # noqa, pylint: disable=unused-import

from pyramid.httpexceptions import HTTPBadRequest
from sqlalchemy.orm.exc import NoResultFound

from c2cgeoportal_commons.models import DBSession
from c2cgeoportal_commons.models.main import OGCServer
from c2cgeoportal_geoportal.lib import get_url2
from c2cgeoportal_geoportal.lib.caching import get_region
from c2cgeoportal_geoportal.views.proxy import Proxy

CACHE_REGION = get_region()
LOG = logging.getLogger(__name__)


class OGCProxy(Proxy):

    params = {}  # type: Dict[str, str]

    def __init__(self, request):
        Proxy.__init__(self, request)

        # params hold the parameters we"re going to send to backend
        self.params = dict(self.request.params)

        # reset possible value of role_id and user_id
        if "role_id" in self.params:  # pragma: no cover
            del self.params["role_id"]
        if "user_id" in self.params:  # pragma: no cover
            del self.params["user_id"]

        self.lower_params = self._get_lower_params(self.params)
        if "ogcserver" not in self.params:
            raise HTTPBadRequest("The querytring argument 'ogcserver' is required")
        self.ogc_server = self._get_ogcserver_byname(self.params["ogcserver"])

    @CACHE_REGION.cache_on_arguments()
    def _get_ogcserver_byname(self, name):
        try:
            result = DBSession.query(OGCServer).filter(OGCServer.name == name).one()
            DBSession.expunge(result)
            return result
        except NoResultFound:  # pragma nocover
            raise HTTPBadRequest("OGSServer '{}' does not exists (existing: {}).".format(
                name,
                ",".join([t[0] for t in DBSession.query(OGCServer.name).all()])
            ))

    def _get_wms_url(self):
        ogc_server = self.ogc_server
        errors = set()
        url = get_url2("The OGC server '{}'".format(ogc_server.name), ogc_server.url, self.request, errors)
        if len(errors) > 0:  # pragma: no cover
            LOG.error("\n".join(errors))
        return url

    def _get_wfs_url(self):
        ogc_server = self.ogc_server
        errors = set()
        url = get_url2(
            "The OGC server (WFS) '{}'".format(ogc_server.name),
            ogc_server.url_wfs or ogc_server.url, self.request, errors
        )
        if len(errors) > 0:  # pragma: no cover
            LOG.error("\n".join(errors))
        return url



# -*- coding: utf-8 -*-

# Copyright (c) 2012-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import inspect
import logging

from dogpile.cache.util import compat
from dogpile.cache.region import make_region

log = logging.getLogger(__name__)
_regions = {}


def map_dbobject(item):
    # here to avoid import loop
    from c2cgeoportal_commons.models import Base

    return item.id if isinstance(item, Base) and hasattr(item, "id") else item


def keygen_function(namespace, function):
    """
    Return a function that generates a string
    key, based on a given function as well as
    arguments to the returned function itself.

    This is used by :meth:`.CacheRegion.cache_on_arguments`
    to generate a cache key from a decorated function.
    """

    if namespace is None:
        namespace = "{0!s}:{1!s}".format(function.__module__, function.__name__)
    else:  # pragma: no cover
        namespace = "{0!s}:{1!s}|{2!s}".format(function.__module__, function.__name__, namespace)

    args = inspect.getfullargspec(function)
    has_self = args[0] and args[0][0] in ("self", "cls")

    def generate_key(*args, **kw):
        if kw:  # pragma: no cover
            raise ValueError(
                "key creation function does not accept keyword arguments.")
        parts = [namespace]
        if has_self:
            args = args[1:]
        parts.extend(map(compat.text_type, map(map_dbobject, args)))
        return "|".join(parts)
    return generate_key


def init_region(conf, region=None):
    """
    Initialize the caching module.
    """
    cache_region = make_region(function_key_generator=keygen_function)
    kwargs = dict(
        (k, conf[k]) for k in
        ("arguments", "expiration_time") if k in conf)
    cache_region.configure(conf["backend"], **kwargs)
    _regions[region] = cache_region
    return cache_region


def get_region(region=None):
    """
    Return a cache region.
    """
    try:
        return _regions[region]
    except KeyError:  # pragma: no cover
        raise Exception(
            "No such caching region. A region must be"
            "initialized before it can be used")


def invalidate_region(region=None):
    return get_region(region).invalidate()


NO_CACHE = 0
PUBLIC_CACHE = 1
PRIVATE_CACHE = 2

CORS_METHODS = "HEAD, GET, POST, PUT, DELETE"


def set_cors_headers(service_headers_settings, request, service_name,
                     credentials, response):
    """
    Handle CORS requests, as specified in http://www.w3.org/TR/cors/
    """
    if "Origin" not in request.headers:
        return  # Not a CORS request if this header is missing
    origin = request.headers["Origin"]

    if not service_headers_settings or \
            "access_control_allow_origin" not in service_headers_settings:
        log.warning("CORS query not configured for service=%s", service_name)
        return

    preflight = request.method == "OPTIONS"
    if preflight and "Access-Control-Request-Method" not in request.headers:
        log.warning("CORS preflight query missing the Access-Control-Request-Method header")
        return

    allowed_origins = service_headers_settings["access_control_allow_origin"]
    if origin not in allowed_origins:
        if "*" in allowed_origins:
            origin = "*"
            credentials = False  # Force no credentials
        else:
            log.warning(
                "CORS query not allowed for origin=%s, service=%s",
                origin, service_name
            )
            return

    response.headers.update({
        "Access-Control-Allow-Origin": origin,
        "Access-Control-Allow-Methods": CORS_METHODS,
    })
    if "Vary" not in response.headers:
        response.headers["Vary"] = set([])
    response.headers["Vary"].add("Origin")

    if preflight:
        max_age = service_headers_settings.get("access_control_max_age", 3600)
        response.headers["Access-Control-Max-Age"] = str(max_age)

    if credentials:
        response.headers["Access-Control-Allow-Credentials"] = "true"

    requested_headers = \
        request.headers.get("Access-Control-Request-Headers", False)
    if requested_headers:
        # For the moment, we allow all requested headers
        response.headers["Access-Control-Allow-Headers"] = requested_headers

    # If we start using headers in responses, we'll have to add
    # Access-Control-Expose-Headers


def set_common_headers(
        request, service_name, cache,
        response=None, credentials=True, vary=False, content_type=None):
    """
    Set the common headers
    vary: Vary on Accept-Language
    """
    if response is None:
        response = request.response

    if cache == NO_CACHE:
        response.cache_control.no_cache = True
        response.cache_control.max_age = 0
    elif cache == PUBLIC_CACHE:
        response.cache_control.public = True
    elif cache == PRIVATE_CACHE:
        if request.user is not None:
            response.cache_control.private = True
        else:
            response.cache_control.public = True
    else:  # pragma: no cover
        raise "Invalid cache type"

    response.headers["Vary"] = set(["Accept-Encoding"])
    if hasattr(request, "registry"):
        headers_settings = request.registry.settings.get("headers", {})
        service_headers_settings = headers_settings.get(service_name, {})

        if cache != NO_CACHE:
            max_age = service_headers_settings.get("cache_control_max_age", 3600)

            response.cache_control.max_age = max_age
            if max_age == 0:
                response.cache_control.no_cache = True

        set_cors_headers(
            service_headers_settings, request, service_name,
            credentials, response
        )

    if vary:
        response.headers["Vary"].add("Accept-Language")

    response.headers["Vary"] = ", ".join(response.headers["Vary"])

    if content_type is not None:
        response.content_type = content_type

    return response



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import argparse
import transaction

from c2cgeoportal_commons.testing import get_session
from c2cgeoportal_geoportal.scripts import fill_arguments, get_appsettings


def main():
    """
    Emergency user create and password reset script
    example, reset toto password to foobar:
    ./docker-compose-run manage_users --password=foobar toto
    example, create user foo with password bar and role admin:
    ./docker-compose-run manage_users --create --rolename=role_admin --password=bar foo

    to get the options list, do:
    ./docker-compose-run manage_users --help
    """

    usage = """Usage: %prog [options] USERNAME

Reset a user password.
The username is used as password if the password is not provided with the corresponding option.
User can be created if it does not exist yet."""

    parser = argparse.ArgumentParser(description=usage)
    fill_arguments(parser)
    parser.add_argument(
        "--password", "-p",
        help="Set password (if not set, username is used as password"
    )
    parser.add_argument(
        "--create", "-c",
        action="store_true", default=False,
        help="Create user if it does not already exist"
    )
    parser.add_argument(
        "--rolename", "-r",
        default="role_admin",
        help="The role name which must exist in the database"
    )
    parser.add_argument(
        "--email", "-e",
        default=None,
        help="The user email"
    )
    parser.add_argument(
        'user',
        help="The user"
    )
    options = parser.parse_args()
    username = options.user
    settings = get_appsettings(options)

    with transaction.manager:
        sess = get_session(settings, transaction.manager)

        # Must be done only once we have loaded the project config
        from c2cgeoportal_commons.models import main, static

        print("\n")

        # Check that user exists
        query = sess.query(static.User).filter_by(username=username)

        result = query.count()
        if result == 0:
            if not options.create:
                # If doesn't exist and no -c option, throw error
                print("User {} does not exist in database".format(username))
                exit(1)
            else:
                if options.password is None:
                    parser.error("The password is mandatory on user creation")
                if options.email is None:
                    parser.error("The email is mandatory on user creation")

                # Get roles
                query_role = sess.query(main.Role).filter(main.Role.name == options.rolename)

                if query_role.count() == 0:
                    # Role not found in db?
                    print("Role matching {} does not exist in database".format(options.rolename))
                    exit(1)

                role = query_role.first()

                user = static.User(
                    username=username,
                    password=options.password,
                    email=options.email,
                    settings_role=role,
                    roles=[role]
                )
                sess.add(user)

                print(("User {} created with password {} and role {}".format(
                    username, options.password, options.rolename
                )))

        else:
            # If user exists (assuming username are unique)
            user = query.first()

            if options.password is not None:
                print("Password set to: {}".format(options.password))
                user.password = "{}".format(options.password)

            if options.email is not None:
                user.email = options.email

            sess.add(user)

            print("Password reset for user {}".format(username))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.
import threading
import warnings
from typing import Dict, Tuple  # noqa, pylint: disable=unused-import

from sqlalchemy import Table, MetaData, Column, Integer
from sqlalchemy.orm import relationship
from sqlalchemy.orm.util import class_mapper
from sqlalchemy.exc import SAWarning
from sqlalchemy.ext.declarative.api import DeclarativeMeta  # noqa, pylint: disable=unused-import

from papyrus.geo_interface import GeoInterface


_class_cache = {}  # type: Dict[Tuple[str, str, str], DeclarativeMeta]

SQL_GEOMETRY_COLUMNS = """
    SELECT srid, type
    FROM geometry_columns
    WHERE
      f_table_schema = :table_schema AND
      f_table_name = :table_name AND
      f_geometry_column = :geometry_column
    """


def init():
    """
    Initialize the db reflection module. Give the declarative base
    class an engine, required for the reflection.
    """
    global _class_cache
    _class_cache.clear()


class _AssociationProxy(object):
    # A specific "association proxy" implementation

    def __init__(self, target, value_attr, nullable=True):
        self.target = target
        self.value_attr = value_attr
        self.nullable = nullable

    def __get__(self, obj, type_=None):
        if obj is None:
            # For "hybrid" descriptors that work both at the instance
            # and class levels we could return an SQL expression here.
            # The code of hybrid_property in SQLAlchemy illustrates
            # how to do that.
            return self
        target = getattr(obj, self.target)
        return getattr(target, self.value_attr) if target else None

    def __set__(self, obj, val):
        from c2cgeoportal_commons.models import DBSession
        o = getattr(obj, self.target)
        # if the obj as no child object or if the child object
        # does not correspond to the new value then we need to
        # read a new child object from the database
        if not o or getattr(o, self.value_attr) != val:
            relationship_property = class_mapper(obj.__class__) \
                .get_property(self.target)
            child_cls = relationship_property.argument
            o = DBSession.query(child_cls).filter(
                getattr(child_cls, self.value_attr) == val).first()
            setattr(obj, self.target, o)


def _get_schema(tablename):
    if "." in tablename:
        schema, tablename = tablename.split(".", 1)
    else:
        schema = "public"

    return tablename, schema


_get_table_lock = threading.RLock()


def get_table(tablename, schema=None, session=None, primary_key=None):
    if schema is None:
        tablename, schema = _get_schema(tablename)

    if session is not None:
        engine = session.bind.engine
        metadata = MetaData(bind=engine)
    else:
        from c2cgeoportal_commons.models import DBSession
        from c2cgeoportal_commons.models.main import Base
        engine = DBSession.bind.engine
        metadata = Base.metadata

    # create table and reflect it
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            "Did not recognize type 'geometry' of column",
            SAWarning)
        args = [tablename, metadata]
        if primary_key is not None:
            # Ensure we have a primary key to be able to edit views
            args.append(Column(primary_key, Integer, primary_key=True))
        with _get_table_lock:
            table = Table(
                *args,
                schema=schema,
                autoload=True,
                autoload_with=engine
            )
    return table


def get_class(tablename, session=None, exclude_properties=None, primary_key=None, attributes_order=None,
              readonly_attributes=None):
    """
    Get the SQLAlchemy mapped class for "tablename". If no class exists
    for "tablename" one is created, and added to the cache. "tablename"
    must reference a valid string. If there is no table identified by
    tablename in the database a NoSuchTableError SQLAlchemy exception
    is raised.
    """

    tablename, schema = _get_schema(tablename)
    cache_key = (
        schema,
        tablename,
        tuple(exclude_properties or ()),
        primary_key,
        tuple(attributes_order or ()),
        tuple(readonly_attributes or ()),
    )

    if cache_key in _class_cache:
        return _class_cache[cache_key]

    table = get_table(tablename, schema, session, primary_key=primary_key)

    # create the mapped class
    cls = _create_class(
        table,
        exclude_properties=exclude_properties,
        attributes_order=attributes_order,
        readonly_attributes=readonly_attributes,
    )

    # add class to cache
    _class_cache[cache_key] = cls

    return cls


def _create_class(
    table, exclude_properties=None, attributes_order=None, readonly_attributes=None, pk_name=None
):
    from c2cgeoportal_commons.models.main import Base

    exclude_properties = exclude_properties or ()
    attributes = dict(
        __table__=table,
        __mapper_args__={"exclude_properties": exclude_properties},
        __attributes_order__=attributes_order,
    )
    if pk_name is not None:
        attributes[pk_name] = Column(Integer, primary_key=True)
    cls = type(
        table.name.capitalize(),
        (GeoInterface, Base),
        attributes,
    )

    for col in table.columns:
        if col.name in (readonly_attributes or []):
            col.info['readonly'] = True
        if col.foreign_keys and col.name not in exclude_properties:
            _add_association_proxy(cls, col)

    return cls


def _add_association_proxy(cls, col):
    if len(col.foreign_keys) != 1:  # pragma: no cover
        raise NotImplementedError

    fk = next(iter(col.foreign_keys))
    child_tablename, child_pk = fk.target_fullname.rsplit(".", 1)
    child_cls = get_class(child_tablename)

    try:
        proxy = col.name[0:col.name.rindex("_id")]
    except ValueError:  # pragma: no cover
        proxy = col.name + "_"

    # The following is a workaround for a bug in the geojson lib. The
    # geojson lib indeed treats properties named "type" specifically
    # (this is a GeoJSON keyword), and produces a UnicodeEncodeError
    # if the property includes non-ascii characters.
    if proxy == "type":
        proxy = "type_"  # pragma: no cover

    rel = proxy + "_"
    primaryjoin = (getattr(cls, col.name) == getattr(child_cls, child_pk))
    relationship_ = relationship(child_cls, primaryjoin=primaryjoin,
                                 lazy="immediate")
    setattr(cls, rel, relationship_)

    nullable = True
    cls_column_property = getattr(cls, col.name).property
    for column in cls_column_property.columns:
        nullable = nullable and column.nullable

    setattr(cls, proxy, _AssociationProxy(rel, "name", nullable=nullable))

    if cls.__add_properties__ is None:
        cls.__add_properties__ = [proxy]
    else:
        cls.__add_properties__.append(proxy)

    col.info['association_proxy'] = proxy



# -*- coding: utf-8 -*-

# Copyright (c) 2012-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


from geojson import loads
from shapely.geometry import asShape
from geoalchemy2.shape import from_shape, to_shape
from pyramid.view import view_config
from sqlalchemy import func
from pyramid.httpexceptions import HTTPBadRequest
from c2cgeoportal_commons.models import DBSession


class GeometryProcessing:

    def __init__(self, request):
        self.request = request

    @view_config(route_name="difference", renderer="geojson")
    def difference(self):
        body = loads(self.request.body)
        if "geometries" not in body or \
                not isinstance(body["geometries"], list) or \
                len(body["geometries"]) != 2:  # pragma: no cover
            raise HTTPBadRequest("""Wrong body, it should be like that:
            {
                "geometries": [geomA, geomB]
            }
            """)

        return to_shape(DBSession.query(func.ST_Difference(
            from_shape(asShape(body["geometries"][0])),
            from_shape(asShape(body["geometries"][1]))
        )).scalar())



# -*- coding: utf-8 -*-

# Copyright (c) 2012-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import logging
import math
import os
from decimal import Decimal
from typing import Dict, Union  # noqa, pylint: disable=unused-import

from fiona.collection import Collection
import rasterio
import zope.event.classhandler
from pyramid.httpexceptions import HTTPBadRequest, HTTPNotFound
from pyramid.view import view_config

from c2cgeoportal_geoportal.lib.caching import NO_CACHE, set_common_headers

log = logging.getLogger(__name__)


class Raster:
    data = {}  # type: Dict[str, Union[Collection, rasterio.DatasetReader]]

    def __init__(self, request):
        self.request = request
        self.rasters = self.request.registry.settings["raster"]

        from c2cgeoportal_commons.models.main import InvalidateCacheEvent

        @zope.event.classhandler.handler(InvalidateCacheEvent)
        def handle(event: InvalidateCacheEvent):  # pylint: disable=unused-variable
            del event
            for _, v in Raster.data.items():
                v.close()
            Raster.data = {}

    def _get_required_finite_float_param(self, name: str) -> float:
        if name not in self.request.params:
            raise HTTPBadRequest("'{}' should be in the query string parameters".format(name))
        try:
            result = float(self.request.params[name])
        except ValueError:
            raise HTTPBadRequest("'{}' ({}) parameters should be a number".format(
                name, self.request.params[name]
            ))
        if not math.isfinite(result):
            raise HTTPBadRequest("'{}' ({}) parameters should be a finite number".format(
                name, self.request.params[name]
            ))
        return result

    @view_config(route_name="raster", renderer="decimaljson")
    def raster(self):
        lon = self._get_required_finite_float_param('lon')
        lat = self._get_required_finite_float_param('lat')

        if "layers" in self.request.params:
            rasters = {}
            layers = self.request.params["layers"].split(",")
            for layer in layers:
                if layer in self.rasters:
                    rasters[layer] = self.rasters[layer]
                else:
                    raise HTTPNotFound("Layer {} not found".format(layer))
        else:
            rasters = self.rasters

        result = {}
        for ref in list(rasters.keys()):
            result[ref] = self._get_raster_value(rasters[ref], ref, lon, lat)

        set_common_headers(self.request, "raster", NO_CACHE)
        return result

    def _get_data(self, layer, name):
        if name not in self.data:
            path = layer["file"]
            if layer.get("type", "shp_index") == "shp_index":
                self.data[name] = Collection(path)
            elif layer.get("type") == "gdal":
                self.data[name] = rasterio.open(path)

        return self.data[name]

    def _get_raster_value(self, layer, name, lon, lat):
        data = self._get_data(layer, name)
        if layer.get("type", "shp_index") == "shp_index":
            tiles = [e for e in data.filter(mask={
                "type": "Point",
                "coordinates": [lon, lat],
            })]

            if len(tiles) == 0:
                return None

            path = os.path.join(
                os.path.dirname(layer["file"]),
                tiles[0]["properties"]["location"],
            )

            with rasterio.open(path) as dataset:
                result = self._get_value(layer, name, dataset, lon, lat)
        elif layer.get("type") == "gdal":
            result = self._get_value(layer, name, data, lon, lat)

        if "round" in layer:
            result = self._round(result, layer["round"])
        elif result is not None:
            result = Decimal(str(result))

        return result

    @staticmethod
    def _get_value(layer, name, dataset, lon, lat):
        index = dataset.index(lon, lat)

        shape = dataset.shape
        if index[0] >= 0 and index[1] >= 0 and index[0] < shape[0] and index[1] < shape[1]:
            def get_index(index_):
                return (index_, index_ + 1)
            result = dataset.read(1, window=(
                get_index(index[0]),
                get_index(index[1]),
            ))[0][0]  # pylint: disable=no-member
            result = None if result == layer.get("nodata", dataset.nodata) else result
        else:
            log.debug(
                "Out of index for layer: %s (%s), "
                "lon/lat: %dx%d, index: %dx%d, shape: %dx%d.",
                name, layer["file"],
                lon, lat, index[0], index[1], dataset.shape[0], dataset.shape[1],
            )
            result = None

        return result

    @staticmethod
    def _round(value, round_to):
        if value is not None:
            return Decimal(str(value)).quantize(Decimal(str(round_to)))
        else:
            return None



# -*- coding: utf-8 -*-

# Copyright (c) 2013-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

# pylint: disable=missing-docstring,attribute-defined-outside-init,protected-access


from unittest import TestCase

from pyramid.testing import DummyRequest

from c2cgeoportal_geoportal.lib.checker import build_url


class TestExportCSVView(TestCase):

    def test_build_url_docker(self):
        request = DummyRequest()
        request.registry.settings = {
            "checker": {
                "base_internal_url": "http://localhost:8080",
            }
        }
        self.assertEqual(
            build_url(
                "Test",
                "/toto?titi#tutu",
                request
            ),
            {
                "url": "http://localhost:8080/toto?titi#tutu",
                "headers": {
                    "Cache-Control": "no-cache",
                }
            }
        )

    def test_build_url_http(self):
        request = DummyRequest()
        request.registry.settings = {
            "checker": {
                "base_internal_url": "http://localhost",
                "forward_host": True,
            }
        }
        self.assertEqual(
            build_url(
                "Test",
                "/toto?titi#tutu",
                request
            ),
            {
                "url": "http://localhost/toto?titi#tutu",
                "headers": {
                    "Cache-Control": "no-cache",
                    "Host": "example.com:80"
                }
            }
        )

    def test_build_url_https(self):
        request = DummyRequest()
        request.registry.settings = {
            "checker": {
                "base_internal_url": "https://localhost",
                "forward_host": True,
            }
        }
        self.assertEqual(
            build_url(
                "Test",
                "/toto?titi#tutu",
                request
            ),
            {
                "url": "https://localhost/toto?titi#tutu",
                "headers": {
                    "Cache-Control": "no-cache",
                    "Host": "example.com:80"
                }
            }
        )

    def test_build_url_forward_headers(self):
        request = DummyRequest()
        request.registry.settings = {
            "checker": {
                "base_internal_url": "http://localhost",
                "forward_headers": ["Cookie"]
            }
        }
        request.headers["Cookie"] = "test"
        self.assertEqual(
            build_url(
                "Test",
                "/toto?titi#tutu",
                request
            ),
            {
                "url": "http://localhost/toto?titi#tutu",
                "headers": {
                    "Cache-Control": "no-cache",
                    "Cookie": "test",
                }
            }
        )



# -*- coding: utf-8 -*-

# Copyright (c) 2014-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import re
import sys
import transaction

from argparse import ArgumentParser
from c2cgeoportal_commons.testing import get_session
from c2cgeoportal_geoportal.scripts import fill_arguments, get_appsettings

from json import loads


def main():
    parser = ArgumentParser(
        prog=sys.argv[0], add_help=True,
        description="Tool used to migrate your old layers from the old structure to the new one.",
    )

    fill_arguments(parser)
    parser.add_argument(
        "--no-layers",
        dest="layers",
        action="store_false",
        help="do not import the layers"
    )
    parser.add_argument(
        "--no-groups",
        dest="groups",
        action="store_false",
        help="do not import the groups"
    )
    parser.add_argument(
        "--no-themes",
        dest="themes",
        action="store_false",
        help="do not import the themes"
    )
    options = parser.parse_args()
    settings = get_appsettings(options)

    with transaction.manager:
        session = get_session(settings, transaction.manager)

        # Must be done only once we have loaded the project config
        from c2cgeoportal_commons.models.main import (
            OGCServer, Theme, LayerWMS, LayerWMTS, LayerV1, LayerGroup)

        if options.layers:
            table_list = [LayerWMTS, LayerWMS, OGCServer]
            for table in table_list:
                print(("Emptying table {0!s}.".format(table.__table__)))
                # Must be done exactly this way otherwise the cascade config in the
                # models are not used
                for t in session.query(table).all():
                    session.delete(t)

            # List and create all distinct ogc_server
            ogc_server(session, settings)

            print("Converting layerv1.")
            for layer in session.query(LayerV1).all():
                layer_v1tov2(session, layer)

        if options.groups:
            print("Converting layer groups.")
            for group in session.query(LayerGroup).all():
                layergroup_v1tov2(session, group)

        if options.themes:
            print("Converting themes.")
            for theme in session.query(Theme).all():
                theme_v1tov2(session, theme)


def ogc_server(session, settings):
    from c2cgeoportal_commons.models.main import LayerV1, OGCServer, \
        OGCSERVER_TYPE_QGISSERVER, OGCSERVER_TYPE_GEOSERVER, OGCSERVER_TYPE_OTHER, \
        OGCSERVER_AUTH_GEOSERVER, OGCSERVER_AUTH_NOAUTH

    qgis_re = []
    geoserver_re = []
    other_re = []
    for e in settings.get("themev1tov2_qgis_res", "").split(","):
        if e.strip() != "":
            qgis_re.append(re.compile(e))
    for e in settings.get("themev1tov2_geoserver_res", "").split(","):
        if e.strip() != "":
            geoserver_re.append(re.compile(e))
    for e in settings.get("themev1tov2_other_res", "").split(","):
        if e.strip() != "":
            other_re.append(re.compile(e))

    servers_v1 = session.query(
        LayerV1.url, LayerV1.image_type, LayerV1.is_single_tile
    ).group_by(
        LayerV1.url, LayerV1.image_type, LayerV1.is_single_tile
    ).filter(LayerV1.layer_type != "WMTS").all()

    # Get existing list of ogc_server
    servers_ogc = session.query(OGCServer).all()
    unique_servers = {
        (
            server.url,
            server.image_type,
            True if server.is_single_tile is None else server.is_single_tile
        )
        for server in servers_ogc
    }

    # Add new ogc_server
    for url, image_type, is_single_tile in servers_v1:
        # Default image_type
        if image_type is None:
            image_type = "image/png"
        if is_single_tile is None:
            is_single_tile = False
        if url is None:
            url = "config://internal/mapserv"
            name = "source for {}".format(image_type)
        else:
            name = "source for {} {}".format(url, image_type)
        if is_single_tile:
            name += " with single_tile"
        identifier = (url, image_type, is_single_tile)
        if identifier not in unique_servers:
            unique_servers.add(identifier)
            new_ogc_server = OGCServer()
            new_ogc_server.url = url
            new_ogc_server.image_type = image_type
            new_ogc_server.is_single_tile = is_single_tile
            new_ogc_server.name = name

            for e in qgis_re:
                if e.search(url) is not None:
                    new_ogc_server.type = OGCSERVER_TYPE_QGISSERVER
                    break
            for e in geoserver_re:
                if e.search(url) is not None:
                    new_ogc_server.type = OGCSERVER_TYPE_GEOSERVER
                    new_ogc_server.auth = OGCSERVER_AUTH_GEOSERVER
                    break
            for e in other_re:
                if e.search(url) is not None:
                    new_ogc_server.type = OGCSERVER_TYPE_OTHER
                    new_ogc_server.auth = OGCSERVER_AUTH_NOAUTH
                    break

            session.add(new_ogc_server)

    session.flush()


def layer_v1tov2(session, layer):
    from c2cgeoportal_commons.models.main import OGCServer, LayerWMS, LayerWMTS, \
        LayergroupTreeitem, Dimension

    if layer.layer_type in ["internal WMS", "external WMS"]:
        # Use the first one
        new_layer = LayerWMS()
        is_single_tile = layer.is_single_tile
        if is_single_tile is None:
            is_single_tile = False
        image_type = "image/png"
        if layer.layer_type == "internal WMS":
            url = "config://internal/mapserv"
        else:
            if layer.image_type is not None:
                image_type = layer.image_type
            url = layer.url
        ogc_server = session.query(OGCServer).filter(
            OGCServer.url == url,
            OGCServer.image_type == image_type,
            OGCServer.is_single_tile == is_single_tile
        ).one()

        new_layer.ogc_server = ogc_server

        new_layer.layer = layer.layer
        new_layer.style = layer.style
        new_layer.time_mode = layer.time_mode
        new_layer.time_widget = layer.time_widget
    elif layer.layer_type == "WMTS":
        new_layer = LayerWMTS()

        new_layer.url = layer.url
        new_layer.layer = layer.layer
        new_layer.style = layer.style
        new_layer.matrix_set = layer.matrix_set
        new_layer.image_type = layer.image_type or "image/png"

        if layer.dimensions is not None:
            dimensions = loads(layer.dimensions)
            for name, value in list(dimensions.items()):
                session.add(Dimension(name, value, new_layer))

    new_layer.name = layer.name
    new_layer.public = layer.public
    new_layer.geo_table = layer.geo_table
    new_layer.interfaces = layer.interfaces
    new_layer.restrictionareas = layer.restrictionareas

    for link in layer.parents_relation:
        new_link = LayergroupTreeitem()
        new_link.ordering = link.ordering
        new_link.description = link.description
        new_link.treegroup = link.treegroup
        new_link.treeitem = new_layer

    layer_add_metadata(layer, new_layer, session)

    session.add(new_layer)


def new_metadata(name, value, item):
    from c2cgeoportal_commons.models.main import Metadata

    metadata = Metadata(name, value)
    metadata.item = item
    return metadata


def layer_add_metadata(layer, new_layer, session):
    if layer.metadata_url is not None:
        session.add(new_metadata("metadataUrl", layer.metadata_url, new_layer))
    if layer.is_checked is True:
        session.add(new_metadata("isChecked", "true", new_layer))
    if layer.icon is not None:
        session.add(new_metadata("iconUrl", layer.icon, new_layer))
    if layer.wms_layers is not None:
        session.add(new_metadata("wmsLayers", layer.wms_layers, new_layer))
    if layer.query_layers is not None:
        session.add(new_metadata("queryLayers", layer.query_layers, new_layer))
    if layer.legend is not None:
        session.add(new_metadata("legend", layer.legend, new_layer))
    if layer.legend_image is not None:
        session.add(new_metadata("legendImage", layer.legend_image, new_layer))
    if layer.legend_rule is not None:
        session.add(new_metadata("legendRule", layer.legend_rule, new_layer))
    if layer.is_legend_expanded is True:
        session.add(new_metadata("isLegendExpanded", "true", new_layer))
    if layer.min_resolution is not None:
        session.add(new_metadata("minResolution", layer.min_resolution, new_layer))
    if layer.max_resolution is not None:
        session.add(new_metadata("maxResolution", layer.max_resolution, new_layer))
    if layer.disclaimer is not None:
        session.add(new_metadata("disclaimer", layer.disclaimer, new_layer))
    if layer.identifier_attribute_field is not None:
        session.add(new_metadata(
            "identifierAttributeField",
            layer.identifier_attribute_field, new_layer
        ))
    if layer.exclude_properties is not None:
        session.add(new_metadata("excludeProperties", layer.exclude_properties, new_layer))


def layergroup_v1tov2(session, group):
    is_expended_metadatas = group.get_metadatas("isExpanded")
    if group.is_expanded is True:
        if len(is_expended_metadatas) > 0:
            is_expended_metadatas[0].value = "true"
        else:
            session.add(new_metadata("isExpanded", "true", group))
    elif len(is_expended_metadatas) > 0:
        session.delete(is_expended_metadatas)


def theme_v1tov2(session, theme):
    thumbnail = theme.get_metadatas("thumbnail")
    if thumbnail is None and theme.icon is not None:
        session.add(new_metadata("thumbnail", theme.icon, theme))



# -*- coding: utf-8 -*-

# Copyright (c) 2014-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import argparse
import transaction

from c2cgeoportal_commons.testing import get_session
from c2cgeoportal_geoportal.scripts import fill_arguments, get_appsettings


def main():
    parser = argparse.ArgumentParser(
        description="This script will rename all the theme elements to removes duplicated elements."
    )
    fill_arguments(parser)
    options = parser.parse_args()
    settings = get_appsettings(options)

    with transaction.manager:
        session = get_session(settings, transaction.manager)

        from c2cgeoportal_commons.models.main import LayerV1, LayerWMS, LayerWMTS, LayerGroup, Theme

        for class_ in [LayerV1, LayerWMS, LayerWMTS, LayerGroup, Theme]:
            names = []
            for item in session.query(class_).all():
                if item.name in names:
                    i = 2
                    while "{}-{}".format(item.name, i) in names:
                        i += 1

                    item.name = "{}-{}".format(item.name, i)
                names.append(item.name)



# -*- coding: utf-8 -*-

from pyramid.security import Allow, ALL_PERMISSIONS


class Root:
    __acl__ = [
        (Allow, 'role_admin', ALL_PERMISSIONS),
        ]

    def __init__(self, request):
        self.request = request



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import argparse
import logging
import transaction

from c2cgeoportal_commons.testing import get_session
from c2cgeoportal_geoportal.scripts import fill_arguments, get_appsettings


LOG = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(
        description="Create and populate the database tables."
    )
    fill_arguments(parser)
    options = parser.parse_args()
    settings = get_appsettings(options)

    with transaction.manager:
        session = get_session(settings, transaction.manager)

        from c2cgeoportal_commons.models.main import Interface, OGCServer, Theme, LayerGroup, LayerWMS

        interfaces = session.query(Interface).all()
        ogc_server = session.query(OGCServer).filter(OGCServer.name == "source for image/png").one()

        layer_borders = LayerWMS("Borders", "borders")
        layer_borders.interfaces = interfaces
        layer_borders.ogc_server = ogc_server
        layer_density = LayerWMS("Density", "density")
        layer_density.interfaces = interfaces
        layer_density.ogc_server = ogc_server

        group = LayerGroup("Demo")
        group.children = [layer_borders, layer_density]

        theme = Theme("Demo")
        theme.children = [group]
        theme.interfaces = interfaces

        session.add(theme)

        print("Successfully added the demo theme")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import subprocess
import os
import yaml
import re
import traceback
import requests
from typing import Dict  # noqa, pylint: disable=unused-import
from json import loads
from urllib.parse import urlsplit
from defusedxml.minidom import parseString
from xml.parsers.expat import ExpatError
from sqlalchemy.exc import ProgrammingError, NoSuchTableError, OperationalError
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.util import class_mapper
from sqlalchemy.orm.properties import ColumnProperty
from geoalchemy2.types import Geometry

from lingua.extractors import Extractor
from lingua.extractors import Message
from pyramid.paster import bootstrap
from bottle import template, MakoTemplate
from mako.template import Template
from mako.lookup import TemplateLookup
from owslib.wms import WebMapService

import c2cgeoportal_geoportal
from c2c.template.config import config
from c2cgeoportal_geoportal import init_dbsessions
from c2cgeoportal_geoportal.lib import add_url_params, get_url2
from c2cgeoportal_geoportal.lib.bashcolor import colorize, RED
from c2cgeoportal_geoportal.lib.caching import init_region
from c2cgeoportal_geoportal.lib.print_ import *  # noqa


class _Registry:  # pragma: no cover
    settings = None

    def __init__(self, settings):
        self.settings = settings


class _Request:  # pragma: no cover
    registry = None  # type: _Registry
    params = {}  # type: Dict[str, str]
    matchdict = {}  # type: Dict[str, str]
    GET = {}  # type: Dict[str, str]
    user_agent = ""  # type: str

    def __init__(self, settings=None):
        self.registry = _Registry(settings)

    @staticmethod
    def static_url(*args, **kwargs):
        del args
        del kwargs
        return ""

    @staticmethod
    def static_path(*args, **kwargs):
        del args
        del kwargs
        return ""

    @staticmethod
    def route_url(*args, **kwargs):
        del args
        del kwargs
        return ""

    @staticmethod
    def current_route_url(*args, **kwargs):
        del args
        del kwargs
        return ""


class GeomapfishAngularExtractor(Extractor):  # pragma: no cover
    """
    GeoMapFish angular extractor
    """

    extensions = [".js", ".html"]

    def __init__(self):
        super().__init__()
        if os.path.exists("geoportal/config.yaml"):
            config.init("geoportal/config.yaml")
            self.config = config.get_config()
        else:
            self.config = None
        self.tpl = None

    def __call__(self, filename, options):

        init_region({"backend": "dogpile.cache.memory"})

        int_filename = filename
        if re.match("^" + re.escape("./{}/templates".format(self.config["package"])), filename):
            try:
                empty_template = Template("")

                class Lookup(TemplateLookup):
                    @staticmethod
                    def get_template(uri):
                        del uri  # unused
                        return empty_template

                class MyTemplate(MakoTemplate):
                    tpl = None

                    def prepare(self, **kwargs):
                        options.update({"input_encoding": self.encoding})
                        lookup = Lookup(**kwargs)
                        if self.source:
                            self.tpl = Template(self.source, lookup=lookup, **kwargs)
                        else:
                            self.tpl = Template(
                                uri=self.name,
                                filename=self.filename,
                                lookup=lookup, **kwargs)

                try:
                    processed = template(
                        filename,
                        {
                            "request": _Request(self.config),
                            "lang": "fr",
                            "debug": False,
                            "extra_params": {},
                            "permalink_themes": "",
                            "fulltextsearch_groups": [],
                            "wfs_types": [],
                            "_": lambda x: x,
                        },
                        template_adapter=MyTemplate
                    )
                    int_filename = os.path.join(os.path.dirname(filename), "_" + os.path.basename(filename))
                    with open(int_filename, "wb") as file_open:
                        file_open.write(processed.encode("utf-8"))
                except Exception:
                    print(colorize(
                        "ERROR! Occurred during the '{}' template generation".format(filename),
                        RED
                    ))
                    print(colorize(traceback.format_exc(), RED))
                    if os.environ.get("IGNORE_I18N_ERRORS", "FALSE") == "TRUE":
                        # Continue with the original one
                        int_filename = filename
                    else:
                        raise
            except Exception:
                print(traceback.format_exc())

        message_str = subprocess.check_output([
            "node", "tools/extract-messages.js", int_filename
        ]).decode("utf-8")
        if int_filename != filename:
            os.unlink(int_filename)
        try:
            messages = []
            for contexts, message in loads(message_str):
                for context in contexts.split(", "):
                    messages.append(Message(
                        None, message, None, [], "", "", context.split(":")
                    ))
            return messages
        except Exception:
            print(colorize("An error occurred", RED))
            print(colorize(message_str, RED))
            print("------")
            raise


class GeomapfishConfigExtractor(Extractor):  # pragma: no cover
    """
    GeoMapFish config extractor (raster layers, and print templates)
    """

    extensions = [".yaml", ".tmpl"]

    def __call__(self, filename, options):
        init_region({"backend": "dogpile.cache.memory"})

        with open(filename) as config_file:
            config = yaml.safe_load(config_file)
            # for application config (config.yaml)
            if "vars" in config:
                return self._collect_app_config(filename)
            # for the print config
            elif "templates" in config:
                return self._collect_print_config(config, filename)
            else:
                raise Exception("Not a known config file")

    def _collect_app_config(self, filename):
        config.init(filename)
        settings = config.get_config()
        # Collect raster layers names
        raster = [
            Message(
                None, raster_layer, None, [], "", "",
                (filename, "raster/{}".format(raster_layer))
            )
            for raster_layer in list(settings.get("raster", {}).keys())
        ]

        # Collect layers enum values (for filters)

        class R:
            settings = None

        class C:
            registry = R()

        config_ = C()
        config_.registry.settings = settings
        init_dbsessions(settings, config_)
        from c2cgeoportal_commons.models import DBSessions
        from c2cgeoportal_commons.models.main import Metadata
        c2cgeoportal_geoportal.init_dbsessions(settings, config_)
        from c2cgeoportal_geoportal.views.layers import Layers
        enums = []
        enum_layers = settings.get("layers", {}).get("enum", {})
        for layername in list(enum_layers.keys()):
            layerinfos = enum_layers.get(layername, {})
            attributes = layerinfos.get("attributes", {})
            for fieldname in list(attributes.keys()):
                values = self._enumerate_attributes_values(DBSessions, Layers, layerinfos, fieldname)
                for value, in values:
                    if value is not None and isinstance(value, str) and value != "":
                        msgid = value
                        location = "/layers/{}/values/{}/{}".format(
                            layername,
                            fieldname,
                            value.encode("ascii", errors="replace") if isinstance(value, str) else value)
                        enums.append(Message(None, msgid, None, [], "", "", (filename, location)))

        metadata_list = []
        defs = config["admin_interface"]["available_metadata"]
        names = [e["name"] for e in defs if e.get("translate", False)]

        if len(names) > 0:
            import sqlalchemy
            engine = sqlalchemy.create_engine(config["sqlalchemy.url"])
            Session = sqlalchemy.orm.session.sessionmaker()  # noqa
            Session.configure(bind=engine)
            session = Session()

            query = session.query(Metadata).filter(Metadata.name.in_(names))
            for metadata in query.all():
                location = "metadata/{}/{}".format(metadata.name, metadata.id)
                metadata_list.append(
                    Message(None, metadata.value, None, [], u"", u"", (filename, location))
                )

        return raster + enums + metadata_list

    @staticmethod
    def _enumerate_attributes_values(dbsessions, layers, layerinfos, fieldname):
        dbname = layerinfos.get("dbsession", "dbsession")
        translate = layerinfos.get("attributes").get(fieldname, {}).get("translate", True)
        if not translate:
            return []
        try:
            dbsession = dbsessions.get(dbname)
            return layers.query_enumerate_attribute_values(dbsession, layerinfos, fieldname)
        except Exception as e:
            table = layerinfos.get("attributes").get(fieldname, {}).get("table")
            print(colorize(
                "ERROR! Unable to collect enumerate attributes for "
                "db: {}, table: {}, column: {} ({})".format(dbname, table, fieldname, e),
                RED
            ))
            if os.environ.get("IGNORE_I18N_ERRORS", "FALSE") == "TRUE":
                return []
            else:
                raise

    @staticmethod
    def _collect_print_config(config, filename):
        result = []
        for template_ in list(config.get("templates").keys()):
            result.append(Message(
                None, template_, None, [], "", "",
                (filename, "template/{}".format(template_))
            ))
            result += [
                Message(
                    None, attribute, None, [], "", "",
                    (filename, "template/{}/{}".format(template_, attribute))
                )
                for attribute in list(config.get("templates")[template_].attributes.keys())
            ]
        return result


class GeomapfishThemeExtractor(Extractor):  # pragma: no cover
    """
    GeoMapFish theme extractor
    """

    # Run on the development.ini file
    extensions = [".ini"]
    featuretype_cache = {}  # type: Dict[str, Dict]
    wmscap_cache = {}  # type: Dict[str, WebMapService]

    def __init__(self):
        super().__init__()
        if os.path.exists("geoportal/config.yaml"):
            config.init("geoportal/config.yaml")
            self.config = config.get_config()
        else:
            self.config = None
        if os.path.exists("project.yaml"):
            with open("project.yaml") as f:
                self.package = yaml.safe_load(f)
        else:
            self.package = None
        self.env = None

    def __call__(self, filename, options):
        messages = []

        try:
            self.env = bootstrap(filename)

            try:
                from c2cgeoportal_commons.models import DBSession
                from c2cgeoportal_commons.models.main import Theme, LayerGroup, \
                    LayerWMS, LayerWMTS, FullTextSearch

                self._import(Theme, messages)
                self._import(LayerGroup, messages)
                self._import(LayerWMS, messages, self._import_layer_wms)
                self._import(LayerWMTS, messages, self._import_layer_wmts)

                for ln, in DBSession.query(FullTextSearch.layer_name).distinct().all():
                    if ln is not None and ln != "":
                        messages.append(Message(
                            None, ln, None, [], "", "",
                            ("fts", ln.encode("ascii", errors="replace"))
                        ))
            except ProgrammingError as e:
                print(colorize(
                    "ERROR! The database is probably not up to date "
                    "(should be ignored when happen during the upgrade)",
                    RED
                ))
                print(colorize(e, RED))
                if os.environ.get("IGNORE_I18N_ERRORS", "FALSE") != "TRUE":
                    raise
        except NoSuchTableError as e:
            print(colorize(
                "ERROR! The schema didn't seem to exists "
                "(should be ignored when happen during the deploy)",
                RED
            ))
            print(colorize(e, RED))
            if os.environ.get("IGNORE_I18N_ERRORS", "FALSE") != "TRUE":
                raise
        except OperationalError as e:
            print(colorize(
                "ERROR! The database didn't seem to exists "
                "(should be ignored when happen during the deploy)",
                RED
            ))
            print(colorize(e, RED))
            if os.environ.get("IGNORE_I18N_ERRORS", "FALSE") != "TRUE":
                raise

        return messages

    @staticmethod
    def _import(object_type, messages, callback=None):
        from c2cgeoportal_commons.models import DBSession

        items = DBSession.query(object_type)
        for item in items:
            messages.append(Message(
                None, item.name, None, [], "", "",
                (item.item_type, item.name.encode("ascii", errors="replace"))
            ))

            if callback is not None:
                callback(item, messages)

    def _import_layer_wms(self, layer, messages):
        server = layer.ogc_server
        url = server.url_wfs or server.url
        if url is None:
            return
        for wms_layer in layer.layer.split(","):
            self._import_layer_attributes(url, wms_layer, layer.item_type, layer.name, messages)
        if layer.geo_table is not None and layer.geo_table != "":
            try:
                from c2cgeoportal_geoportal.views.layers import get_layer_class
                cls = get_layer_class(layer, with_last_update_columns=True)
                for column_property in class_mapper(cls).iterate_properties:
                    if isinstance(column_property, ColumnProperty) and len(column_property.columns) == 1:
                        column = column_property.columns[0]
                        if not column.primary_key and not isinstance(column.type, Geometry):
                            if column.foreign_keys:
                                if column.name == "type_id":
                                    name = "type_"
                                elif column.name.endswith("_id"):
                                    name = column.name[:-3]
                                else:
                                    name = column.name + "_"
                            else:
                                name = column_property.key
                            messages.append(Message(
                                None, name, None, [], "", "",
                                (".".join(["edit", layer.item_type, str(layer.id)]), layer.name)
                            ))
            except NoSuchTableError:
                print(colorize(
                    "ERROR! No such table '{}' for layer '{}'.".format(layer.geo_table, layer.name),
                    RED
                ))
                print(colorize(traceback.format_exc(), RED))
                if os.environ.get("IGNORE_I18N_ERRORS", "FALSE") != "TRUE":
                    raise

    def _import_layer_wmts(self, layer, messages):
        from c2cgeoportal_commons.models import DBSession
        from c2cgeoportal_commons.models.main import OGCServer

        layers = [d.value for d in layer.metadatas if d.name == "queryLayers"]
        if len(layers) == 0:
            layers = [d.value for d in layer.metadatas if d.name == "wmsLayer"]
        server = [d.value for d in layer.metadatas if d.name == "ogcServer"]
        if len(server) >= 1 and len(layers) >= 1:
            for wms_layer in layers:
                try:
                    db_server = DBSession.query(OGCServer).filter(OGCServer.name == server[0]).one()
                    self._import_layer_attributes(
                        db_server.url_wfs or db_server.url, wms_layer,
                        layer.item_type, layer.name, messages
                    )
                except NoResultFound:
                    print(colorize(
                        "ERROR! the OGC server '{}' from the WMTS layer '{}' does not exist.".format(
                            server[0], layer.name
                        ),
                        RED
                    ))
                    if os.environ.get("IGNORE_I18N_ERRORS", "FALSE") != "TRUE":
                        raise

    def _import_layer_attributes(self, url, layer, item_type, name, messages):
        attributes, layers = self._layer_attributes(url, layer)
        for layer in layers:
            messages.append(Message(
                None, layer, None, [], "", "",
                (".".join([item_type, name]), layer.encode("ascii", "replace"))
            ))
        for attribute in attributes:
            messages.append(Message(
                None, attribute, None, [], "", "",
                (".".join([item_type, name]), layer.encode("ascii", "replace"))
            ))

    def _build_url(self, url):
        url_split = urlsplit(url)
        hostname = url_split.hostname
        host_map = self.config.get('lingua_extractor', {}).get('host_map', {})
        if hostname in host_map:
            map_ = host_map[hostname]
            if 'netloc' in map_:
                url_split = url_split._replace(netloc=map_['netloc'])
            if 'scheme' in map_:
                url_split = url_split._replace(scheme=map_['scheme'])
            kwargs = {'verify': map_['verify']} if 'verify' in map_ else {}
            return url_split.geturl(), map_.get('headers', {}), kwargs
        return url, {}, {}

    def _layer_attributes(self, url, layer):
        errors = set()

        request = _Request()
        request.registry.settings = self.config
        # static schema will not be supported
        url = get_url2("Layer", url, request, errors)
        if len(errors) > 0:
            print("\n".join(errors))
            return [], []
        url, headers, kwargs = self._build_url(url)

        if url not in self.wmscap_cache:
            print("Get WMS GetCapabilities for URL: {}".format(url))
            self.wmscap_cache[url] = None

            wms_getcap_url = add_url_params(url, {
                "SERVICE": "WMS",
                "VERSION": "1.1.1",
                "REQUEST": "GetCapabilities",
            })
            try:
                print("Get WMS GetCapabilities for URL {},\nwith headers: {}".format(
                    wms_getcap_url, " ".join(["=".join(h) for h in headers.items()])
                ))
                response = requests.get(wms_getcap_url, headers=headers, **kwargs)

                try:
                    self.wmscap_cache[url] = WebMapService(None, xml=response.content)
                except Exception as e:
                    print(colorize(
                        "ERROR! an error occurred while trying to "
                        "parse the GetCapabilities document.",
                        RED))
                    print(colorize(str(e), RED))
                    print("URL: {}\nxml:\n{}".format(wms_getcap_url, response.text))
                    if os.environ.get("IGNORE_I18N_ERRORS", "FALSE") != "TRUE":
                        raise
            except Exception as e:  # pragma: no cover
                print(colorize(str(e), RED))
                print(colorize(
                    "ERROR! Unable to GetCapabilities from URL: {},\nwith headers: {}".format(
                        wms_getcap_url, " ".join(["=".join(h) for h in headers.items()])
                    ),
                    RED,
                ))
                if os.environ.get("IGNORE_I18N_ERRORS", "FALSE") != "TRUE":
                    raise

        wmscap = self.wmscap_cache[url]

        if url not in self.featuretype_cache:
            print("Get WFS DescribeFeatureType for URL: {}".format(url))
            self.featuretype_cache[url] = None

            wfs_descrfeat_url = add_url_params(url, {
                "SERVICE": "WFS",
                "VERSION": "1.1.0",
                "REQUEST": "DescribeFeatureType",
            })
            try:
                response = requests.get(wfs_descrfeat_url, headers=headers, **kwargs)
            except Exception as e:  # pragma: no cover
                print(colorize(str(e), RED))
                print(colorize(
                    "ERROR! Unable to DescribeFeatureType from URL: {}".format(wfs_descrfeat_url),
                    RED,
                ))
                if os.environ.get("IGNORE_I18N_ERRORS", "FALSE") == "TRUE":
                    return [], []
                else:
                    raise

            if not response.ok:  # pragma: no cover
                print(colorize(
                    "ERROR! DescribeFeatureType from URL {} return the error: {:d} {}".format(
                        wfs_descrfeat_url, response.status_code, response.reason
                    ),
                    RED,
                ))
                if os.environ.get("IGNORE_I18N_ERRORS", "FALSE") == "TRUE":
                    return [], []
                else:
                    raise Exception("Aborted")

            try:
                describe = parseString(response.text)
                featurestype = {}
                self.featuretype_cache[url] = featurestype
                for type_element in describe.getElementsByTagNameNS(
                    "http://www.w3.org/2001/XMLSchema", "complexType"
                ):
                    featurestype[type_element.getAttribute("name")] = type_element
            except ExpatError as e:
                print(colorize(
                    "ERROR! an error occurred while trying to "
                    "parse the DescribeFeatureType document.",
                    RED
                ))
                print(colorize(str(e), RED))
                print("URL: {}\nxml:\n{}".format(wfs_descrfeat_url, response.text))
                if os.environ.get("IGNORE_I18N_ERRORS", "FALSE") == "TRUE":
                    return [], []
                else:
                    raise
            except AttributeError:
                print(colorize(
                    "ERROR! an error occurred while trying to "
                    "read the Mapfile and recover the themes.",
                    RED
                ))
                print("URL: {}\nxml:\n{}".format(wfs_descrfeat_url, response.text))
                if os.environ.get("IGNORE_I18N_ERRORS", "FALSE") == "TRUE":
                    return [], []
                else:
                    raise
        else:
            featurestype = self.featuretype_cache[url]

        if featurestype is None:
            return [], []

        layers = [layer]
        if wmscap is not None and layer in list(wmscap.contents):
            layer_obj = wmscap[layer]
            if len(layer_obj.layers) > 0:
                layers = [l.name for l in layer_obj.layers]

        attributes = []
        for sub_layer in layers:
            # Should probably be adapted for other king of servers
            type_element = featurestype.get("{}Type".format(sub_layer))
            if type_element is not None:
                for element in type_element.getElementsByTagNameNS(
                    "http://www.w3.org/2001/XMLSchema", "element"
                ):
                    if not element.getAttribute("type").startswith("gml:"):
                        attributes.append(element.getAttribute("name"))

        return attributes, layers
